/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: MatrixOperators.h,v 1.31 2003/12/05 19:37:19 opetzold Exp $
 */

#ifndef TVMET_MATRIX_OPERATORS_H
#define TVMET_MATRIX_OPERATORS_H

namespace tvmet {


/**
 * \fn operator<<(std::ostream& os, const Matrix<T, Rows, Cols>& rhs)
 * \brief Overload operator for i/o
 * \ingroup _binary_operator
 */
template<class T, std::size_t Rows, std::size_t Cols>
inline
std::ostream& operator<<(std::ostream& os, const Matrix<T, Rows, Cols>& rhs) {
  return rhs.print_on(os);
}


/***********************************************************************
 * Member operators (arithmetic and bit ops)
 ***********************************************************************/


/*
 * update_operator(Matrix<T1, Rows, Cols>, Matrix<T2, Rows, Cols>)
 * update_operator(Matrix<T1, Rows, Cols>, XprMatrix<E, Rows, Cols> rhs)
 * Note: per se element wise
 * \todo: the operator*= can have element wise mul oder product, decide!
 */
#define TVMET_IMPLEMENT_MACRO(MEMBER, OP)				\
template<class T1, class T2, std::size_t Rows, std::size_t Cols>	\
inline Matrix<T1, Rows, Cols>& 						\
operator OP (Matrix<T1, Rows, Cols>& lhs, const Matrix<T2, Rows, Cols>& rhs) { \
  return lhs.MEMBER(rhs);						\
}									\
									\
template<class T, class E, std::size_t Rows,  std::size_t Cols>		\
inline Matrix<T, Rows, Cols>&						\
operator OP (Matrix<T, Rows, Cols>& lhs, const XprMatrix<E, Rows, Cols>& rhs) { \
  return lhs.MEMBER(rhs);                                              	\
}

TVMET_IMPLEMENT_MACRO(AddUpd, +=)		// per se element wise
TVMET_IMPLEMENT_MACRO(SubUpd, -=)		// per se element wise

namespace element_wise {
  TVMET_IMPLEMENT_MACRO(MulUpd, *=)		// see note
  TVMET_IMPLEMENT_MACRO(DivUpd, /=)		// not defined for vectors
}

// integer operators only, e.g used on double you wil get an error
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(ModUpd, %=)
  TVMET_IMPLEMENT_MACRO(BitwiseXorUpd, ^=)
  TVMET_IMPLEMENT_MACRO(BitwiseAndUpd, &=)
  TVMET_IMPLEMENT_MACRO(BitwiseOrUpd, |=)
  TVMET_IMPLEMENT_MACRO(ShlUpd, <<=)
  TVMET_IMPLEMENT_MACRO(ShrUpd, >>=)
}

#undef TVMET_IMPLEMENT_MACRO


/************************************************************************
 * Matrix arithmetic operators implemented by functions
 * add, sub, mul and div
 ************************************************************************/


/*
 * operator(Matrix<T1, Rows, Cols>, Matrix<T2, Rows, Cols>)
 * operator(XprMatrix<E, Rows, Cols>, Matrix<T, Rows, Cols>)
 * operator(Matrix<T, Rows, Cols>, XprMatrix<E, Rows, Cols>)
 * Note: per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, FUNC, OP)                   	\
template<class T1, class T2, std::size_t Rows, std::size_t Cols>	\
inline									\
XprMatrix<								\
  XprBinOp<								\
    FUNCTIONAL<T1, T2>,							\
    MatrixConstReference<T1, Rows, Cols>,				\
    MatrixConstReference<T2, Rows, Cols>				\
  >, 									\
  Rows, Cols 								\
>									\
operator OP (const Matrix<T1, Rows, Cols>& lhs,	const Matrix<T2, Rows, Cols>& rhs) { \
  return FUNC(lhs, rhs);						\
}                   	      						\
                   	      						\
template<class E, class T, std::size_t Rows, std::size_t Cols>    	\
inline                                 					\
XprMatrix<                                     				\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,				\
    XprMatrix<E, Rows, Cols>,                  				\
    MatrixConstReference<T, Rows, Cols>                			\
  >,                     						\
  Rows, Cols 								\
>                      							\
operator OP (const XprMatrix<E, Rows, Cols>& lhs, const Matrix<T, Rows, Cols>& rhs) { \
  return FUNC(lhs, rhs);						\
}                    	      						\
                    	      						\
template<class T, class E, std::size_t Rows, std::size_t Cols>		\
inline                 			      			      	\
XprMatrix<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,				\
    MatrixConstReference<T, Rows, Cols>,               			\
    XprMatrix<E, Rows, Cols>                   				\
  >,                     	      					\
  Rows, Cols 								\
>                      	      						\
operator OP (const Matrix<T, Rows, Cols>& lhs, const XprMatrix<E, Rows, Cols>& rhs) { \
  return FUNC(lhs, rhs);						\
}

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +)		// per se element wise
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -)		// per se element wise

namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *)	// see as prod()
  TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /)	// not defined for matrizes
}
#undef TVMET_IMPLEMENT_MACRO


/*
 * operator(Matrix<T, Rows, Cols>, POD)
 * operator(POD, Matrix<T, Rows, Cols>)
 * Note: operations +,-,*,/ are per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, FUNC, OP, POD)               	\
template<class T, std::size_t Rows, std::size_t Cols>   	  	\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<T, POD >,          					\
    MatrixConstReference<T, Rows, Cols>,       				\
    XprLiteral<POD >          						\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (const Matrix<T, Rows, Cols>& lhs, POD rhs) {		\
  return FUNC (lhs, rhs);						\
}          								\
          								\
template<class T, std::size_t Rows, std::size_t Cols>  	    		\
inline                 			      			      	\
XprMatrix<     								\
  XprBinOp<          							\
    FUNCTIONAL< POD, T>,          					\
    XprLiteral< POD >,         						\
    MatrixConstReference<T, Rows, Cols>        				\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (POD lhs, const Matrix<T, Rows, Cols>& rhs) {    	  	\
  return FUNC (lhs, rhs);						\
}

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, int)

#if defined(TVMET_HAVE_LONG_LONG)
TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, long long int)
#endif // defined(TVMET_HAVE_LONG_LONG)

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, float)

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, double)

#if defined(TVMET_HAVE_LONG_DOUBLE)
TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, long double)
#endif // defined(TVMET_HAVE_LONG_DOUBLE)

#undef TVMET_IMPLEMENT_MACRO


#if defined(TVMET_HAVE_COMPLEX)
/*
 * operator(Matrix<T, Rows, Cols>, complex<T>)
 * operator(complex<T>, Matrix<T, Rows, Cols>)
 * Note: operations +,-,*,/ are per se element wise
 * \todo type promotion
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, FUNC, OP)	               	\
template<class T, std::size_t Rows, std::size_t Cols>         		\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    MatrixConstReference< std::complex<T>, Rows, Cols>,			\
    XprLiteral<std::complex<T> > 					\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (const Matrix< std::complex<T>, Rows, Cols>& lhs, 		\
	     const std::complex<T>& rhs) { 				\
  return FUNC (lhs, rhs);						\
}          								\
          								\
template<class T, std::size_t Rows, std::size_t Cols>          		\
inline                 			      			      	\
XprMatrix<     								\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    XprLiteral< std::complex<T> >,  					\
    MatrixConstReference< std::complex<T>, Rows, Cols>			\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (const std::complex<T>& lhs, 				\
	     const Matrix< std::complex<T>, Rows, Cols>& rhs) { 	\
  return FUNC (lhs, rhs);						\
}

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /)

#undef TVMET_IMPLEMENT_MACRO

#endif // defined(TVMET_HAVE_COMPLEX)


/************************************************************************
 * matrix specific operator*() = prod() operations
 ************************************************************************/


/**
 * \fn operator*(const Matrix<T1, Rows1, Cols1>& lhs, const Matrix<T2, Cols1, Cols2>& rhs)
 * \brief Multiply two Matrices.
 * \ingroup _binary_operator
 * \note The rows2 has to be equal to cols1.
 * \sa prod(const Matrix<T1, Rows1, Cols1>& lhs, const Matrix<T2, Cols1, Cols2>& rhs)
 */
template<class T1, std::size_t Rows1, std::size_t Cols1,
	 class T2, std::size_t Cols2>
inline
XprMatrix<
  XprMMProduct<
    MatrixConstReference<T1, Rows1, Cols1>, Rows1, Cols1,
    MatrixConstReference<T2, Cols1, Cols2>, Cols2,
    Cols1, 1,
    Cols2, 1
  >,
  Rows1, Cols2
>
operator*(const Matrix<T1, Rows1, Cols1>& lhs, const Matrix<T2, Cols1, Cols2>& rhs) {
  return prod(lhs, rhs);
}


/**
 * \fn operator*(const XprMatrix<E1, Rows1, Cols1>& lhs, const Matrix<T2, Cols1, Cols2>& rhs)
 * \brief Evaluate the product of XprMatrix and Matrix.
 * \ingroup _binary_operator
 * \sa prod(const XprMatrix<E1, Rows1, Cols1>& lhs, const Matrix<T2, Cols1, Cols2>& rhs)
 */
template<class E1, std::size_t Rows1, std::size_t Cols1,
	 class T2, std::size_t Cols2>
inline
XprMatrix<
  XprMMProduct<
    XprMatrix<E1, Rows1, Cols1>, Rows1, Cols1,
    MatrixConstReference<T2, Cols1, Cols2>, Cols2,
    Cols1, 1,
    Cols2, 1
  >,
  Rows1, Cols2
>
operator*(const XprMatrix<E1, Rows1, Cols1>& lhs, const Matrix<T2, Cols1, Cols2>& rhs) {
  return prod(lhs, rhs);
}


/**
 * \fn operator*(const Matrix<T1, Rows1, Cols1>& lhs, const XprMatrix<E2, Cols1, Cols2>& rhs)
 * \brief Evaluate the product of Matrix and XprMatrix.
 * \ingroup _binary_operator
 * \sa prod(const Matrix<T, Rows1, Cols1>& lhs, const XprMatrix<E, Cols1, Cols2>& rhs)
 */
template<class T1, std::size_t Rows1, std::size_t Cols1,
	 class E2, std::size_t Cols2>
inline
XprMatrix<
  XprMMProduct<
    MatrixConstReference<T1, Rows1, Cols1>, Rows1, Cols1,
    XprMatrix<E2, Cols1, Cols2>, Cols2,
    Cols1, 1,
    Cols2, 1
  >,
  Rows1, Cols2
>
operator*(const Matrix<T1, Rows1, Cols1>& lhs, const XprMatrix<E2, Cols1, Cols2>& rhs) {
  return prod(lhs, rhs);
}


/************************************************************************
 * matrix-vector specific prod( ... ) operators
 ************************************************************************/


/**
 * \fn operator*(const Matrix<T1, Rows, Cols>& lhs, const Vector<T2, Cols>& rhs)
 * \brief Multiply a Matrix with a Vector.
 * \ingroup _binary_operator
 * \note The length of the Vector has to be equal to the number of Columns.
 * \sa prod(const Matrix<T1, Rows, Cols>& m, const Vector<T2, Cols>& v)
 */
template<class T1, std::size_t Rows, std::size_t Cols, class T2>
inline
XprVector<
  XprMVProduct<
    MatrixConstReference<T1, Rows, Cols>, Rows, Cols,
    Cols, 1,
    VectorConstReference<T2, Cols, 1>, 1
  >,
  Rows
>
operator*(const Matrix<T1, Rows, Cols>& lhs, const Vector<T2, Cols>& rhs) {
  return prod(lhs, rhs);
}


/**
 * \fn operator*(const Matrix<T1, Rows, Cols>& lhs, const XprVector<E2, Cols>& rhs)
 * \brief Function for the matrix-vector-product
 * \ingroup _binary_operator
 * \sa prod(const Matrix<T, Rows, Cols>& lhs, const XprVector<E, Cols>& rhs)
 */
template<class T1, class E2, std::size_t Rows, std::size_t Cols>
inline
XprVector<
  XprMVProduct<
    MatrixConstReference<T1, Rows, Cols>, Rows, Cols,
    Cols, 1,
    XprVector<E2, Cols>, 1
  >,
  Rows
>
operator*(const Matrix<T1, Rows, Cols>& lhs, const XprVector<E2, Cols>& rhs) {
  return prod(lhs, rhs);
}


/**
 * \fn operator*(const XprMatrix<E1, Rows, Cols>& lhs, const Vector<T2, Cols>& rhs)
 * \brief Compute the product of an XprMatrix with a Vector.
 * \ingroup _binary_operator
 * \sa prod(const XprMatrix<E, Rows, Cols>& lhs, const Vector<T, Cols>& rhs)
 */
template<class E1, class T2, std::size_t Rows, std::size_t Cols>
inline
XprVector<
  XprMVProduct<
    XprMatrix<E1, Rows, Cols>, Rows, Cols,
    Cols, 1,
    VectorConstReference<T2, Cols, 1>, 1
  >,
  Rows
>
operator*(const XprMatrix<E1, Rows, Cols>& lhs, const Vector<T2, Cols>& rhs) {
  return prod(lhs, rhs);
}


/************************************************************************
 * Matrix integer and compare operators
 ************************************************************************/


/*
 * operator(Matrix<T1, Rows, Cols>, Matrix<T2, Rows, Cols>)
 * operator(XprMatrix<E>, Matrix<T, Rows, Cols>)
 * operator(Matrix<T, Rows, Cols>, XprMatrix<E>)
 * Note: operations are per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP)                   	\
template<class T1, std::size_t Rows, std::size_t Cols,			\
	 class T2>							\
inline									\
XprMatrix<								\
  XprBinOp<								\
    FUNCTIONAL<T1, T2>,							\
    MatrixConstReference<T1, Rows, Cols>,				\
    MatrixConstReference<T2, Rows, Cols>				\
  >, 									\
  Rows, Cols 								\
>									\
operator OP (const Matrix<T1, Rows, Cols>& lhs,				\
	     const Matrix<T2, Rows, Cols>& rhs) {			\
  typedef XprBinOp <							\
    FUNCTIONAL<T1, T2>,							\
    MatrixConstReference<T1, Rows, Cols>,				\
    MatrixConstReference<T2, Rows, Cols>				\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(lhs.const_ref(), rhs.const_ref())); \
}									\
									\
template<class E,                   					\
	 class T, std::size_t Rows, std::size_t Cols>         		\
inline                                 					\
XprMatrix<                                     				\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,				\
    XprMatrix<E, Rows, Cols>,                  				\
    MatrixConstReference<T, Rows, Cols>                			\
  >,                     						\
  Rows, Cols 								\
>                      							\
operator OP (const XprMatrix<E, Rows, Cols>& lhs, const Matrix<T, Rows, Cols>& rhs) { \
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<typename E::value_type, T>,				\
    XprMatrix<E, Rows, Cols>,						\
    MatrixConstReference<T, Rows, Cols>					\
  > 							 expr_type; 	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(lhs, rhs.const_ref())); \
}                    	      						\
                    	      						\
template<class T, std::size_t Rows, std::size_t Cols, 			\
	 class E>			                   		\
inline                 			      			      	\
XprMatrix<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,				\
    MatrixConstReference<T, Rows, Cols>,               			\
    XprMatrix<E, Rows, Cols>                   				\
  >,                     	      					\
  Rows, Cols 								\
>                      	      						\
operator OP (const Matrix<T, Rows, Cols>& lhs, const XprMatrix<E, Rows, Cols>& rhs) { \
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<T, typename E::value_type>,				\
    MatrixConstReference<T, Rows, Cols>,               			\
    XprMatrix<E, Rows, Cols>						\
  >	 						 expr_type; 	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(lhs.const_ref(), rhs)); \
}

// integer operators only, e.g used on double you wil get an error
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>)
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||)

#undef TVMET_IMPLEMENT_MACRO


/*
 * operator(Matrix<T, Rows, Cols>, POD)
 * operator(POD, Matrix<T, Rows, Cols>)
 * Note: operations are per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP, TP)               	\
template<class T, std::size_t Rows, std::size_t Cols>         		\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<T, TP >,          					\
    MatrixConstReference<T, Rows, Cols>,       				\
    XprLiteral<TP >          						\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (const Matrix<T, Rows, Cols>& lhs, TP rhs) {		\
  typedef XprBinOp<          						\
    FUNCTIONAL<T, TP >,          					\
    MatrixConstReference<T, Rows, Cols>,       				\
    XprLiteral< TP >							\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(				\
    expr_type(lhs.const_ref(), XprLiteral< TP >(rhs))); 		\
}          								\
          								\
template<class T, std::size_t Rows, std::size_t Cols>          		\
inline                 			      			      	\
XprMatrix<     								\
  XprBinOp<          							\
    FUNCTIONAL< TP, T>,          					\
    XprLiteral< TP >,         						\
    MatrixConstReference<T, Rows, Cols>        				\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (TP lhs, const Matrix<T, Rows, Cols>& rhs) {      		\
  typedef XprBinOp<         						\
    FUNCTIONAL< TP, T>,         					\
    XprLiteral< TP >,         						\
    MatrixConstReference<T, Rows, Cols>					\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(				\
    expr_type(XprLiteral< TP >(lhs), rhs.const_ref())); 		\
}

// integer operators only, e.g used on double you wil get an error
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>, int)
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, int)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, int)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, int)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, int)

#if defined(TVMET_HAVE_LONG_LONG)
// integer operators only
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>, long long int)
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, long long int)
#endif // defined(TVMET_HAVE_LONG_LONG)

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, float)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, float)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, float)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, float)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, float)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, float)

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, double)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, double)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, double)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, double)

#if defined(TVMET_HAVE_LONG_DOUBLE)
// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, long double)
#endif // defined(TVMET_HAVE_LONG_DOUBLE)

#undef TVMET_IMPLEMENT_MACRO


#if defined(TVMET_HAVE_COMPLEX)
/*
 * operator(Matrix<T, Rows, Cols>, complex<T>)
 * operator(complex<T>, Matrix<T, Rows, Cols>)
 * Note: - per se element wise
 *       - bit ops on complex<int> doesn't make sense, stay away
 * \todo type promotion
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP)		               	\
template<class T, std::size_t Rows, std::size_t Cols>         		\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    MatrixConstReference< std::complex<T>, Rows, Cols>,			\
    XprLiteral<std::complex<T> > 					\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (const Matrix< std::complex<T>, Rows, Cols>& lhs, 		\
	     const std::complex<T>& rhs) { 				\
  typedef XprBinOp<          						\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    MatrixConstReference< std::complex<T>, Rows, Cols>,			\
    XprLiteral< std::complex<T> >					\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(				\
    expr_type(lhs.const_ref(), XprLiteral< std::complex<T> >(rhs))); 	\
}          								\
          								\
template<class T, std::size_t Rows, std::size_t Cols>          		\
inline                 			      			      	\
XprMatrix<     								\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    XprLiteral< std::complex<T> >,  					\
    MatrixConstReference< std::complex<T>, Rows, Cols>			\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (const std::complex<T>& lhs, 				\
	     const Matrix< std::complex<T>, Rows, Cols>& rhs) { 	\
  typedef XprBinOp<         						\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    XprLiteral< std::complex<T> >, 					\
    MatrixConstReference<T, Rows, Cols>					\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(				\
    expr_type(XprLiteral< std::complex<T> >(lhs), rhs.const_ref())); 	\
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||)

#undef TVMET_IMPLEMENT_MACRO

#endif // defined(TVMET_HAVE_COMPLEX)


/***********************************************************************
 * global unary operators
 ***********************************************************************/


/*
 * unary_operator(Matrix<T, Rows, Cols>)
 * Note: per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP)                  		\
template <class T, std::size_t Rows, std::size_t Cols>                  \
inline                                                        		\
XprMatrix<                                                       	\
  XprUnOp<                                                        	\
    FUNCTIONAL<T>,                                                      \
    MatrixConstReference<T, Rows, Cols>                                	\
  >,                                                        		\
  Rows, Cols 								\
>                                                        		\
operator OP (const Matrix<T, Rows, Cols>& rhs) {                        \
  typedef XprUnOp<                                                      \
    FUNCTIONAL<T>,                                                      \
    MatrixConstReference<T, Rows, Cols>                                	\
  >  							 expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(rhs.const_ref()));	\
}

TVMET_IMPLEMENT_MACRO(Fcnl_LogicalNot, !)
TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseNot, ~)
TVMET_IMPLEMENT_MACRO(Fcnl_Negate, -)
#undef TVMET_IMPLEMENT_MACRO


} // namespace tvmet

#endif // TVMET_MATRIX_OPERATORS_H

// Local Variables:
// mode:C++
// End:
