/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Matrix.h,v 1.11 2003/11/30 08:26:25 opetzold Exp $
 */

#ifndef TVMET_META_MATRIX_H
#define TVMET_META_MATRIX_H

#include <tvmet/xpr/Null.h>

namespace tvmet {

namespace meta {


/**
 * \class Matrix Matrix.h "tvmet/meta/Matrix.h"
 * \brief Meta %Matrix class using expression and meta templates.
 */
template<std::size_t Rows, std::size_t Cols,
	 std::size_t RowStride=0, std::size_t ColStride=0>
class Matrix
{
  Matrix();
  Matrix(const Matrix&);
  Matrix& operator=(const Matrix&);

private:
  enum {
    doRows = (RowStride < Rows - 1) ? 1 : 0,	/**< recursive counter Rows. */
    doCols = (ColStride < Cols - 1) ? 1 : 0	/**< recursive counter Cols. */
  };

public:
  /** Assign an expression on columns on given row using the functional fn. */
  template<class E1, class E2, class Fcnl>
  static inline
  void assign2(E1& lhs, const E2& rhs, const Fcnl& fn) {
    fn.apply_on(lhs(RowStride, ColStride), rhs(RowStride, ColStride));
    Matrix<Rows * doCols, Cols * doCols,
           RowStride * doCols, (ColStride+1) * doCols>::assign2(lhs, rhs, fn);
  }

  /** Assign an expression on row-wise using the functional fn. */
  template<class E1, class E2, class Fcnl>
  static inline
  void assign(E1& lhs, const E2& rhs, const Fcnl& fn) {
    Matrix<Rows, Cols,
           RowStride, 0>::assign2(lhs, rhs, fn);
    Matrix<Rows * doRows, Cols * doRows,
          (RowStride+1) * doRows, 0>::assign(lhs, rhs, fn);
  }

  /** evaluate a given matrix expression, column wise. */
  template<class E>
  static inline
  bool all_elements2(const E& e) {
    if(!e(RowStride, ColStride)) return false;
    return Matrix<Rows * doCols, Cols * doCols,
                  RowStride * doCols, (ColStride+1) * doCols>::all_elements2(e);
  }

  /** evaluate a given matrix expression, row wise. */
  template<class E>
  static inline
  bool all_elements(const E& e) {
    if(!Matrix<Rows, Cols, RowStride, 0>::all_elements2(e) ) return false;
    return Matrix<Rows * doRows, Cols * doRows,
                 (RowStride+1) * doRows, 0>::all_elements(e);
  }

  /** evaluate a given matrix expression, column wise. */
  template<class E>
  static inline
  bool any_elements2(const E& e) {
    if(e(RowStride, ColStride)) return true;
    return Matrix<Rows * doCols, Cols * doCols,
                  RowStride * doCols, (ColStride+1) * doCols>::any_elements2(e);
  }

  /** evaluate a given matrix expression, row wise. */
  template<class E>
  static inline
  bool any_elements(const E& e) {
    if(Matrix<Rows, Cols, RowStride, 0>::any_elements2(e) ) return true;
    return Matrix<Rows * doRows, Cols * doRows,
                 (RowStride+1) * doRows, 0>::any_elements(e);
  }

  /** trace a given matrix expression. */
  template<class E>
  static inline
  typename E::value_type
  trace(const E& e) {
    return e(RowStride, ColStride)
      + Matrix<Rows * doCols, Cols * doCols,
              (RowStride+1) * doCols, (ColStride+1) * doCols>::trace(e);
  }

};


/**
 * \class Matrix<0, 0, 0, 0> Matrix.h "tvmet/meta/Matrix.h"
 * \brief Meta %Matrix specialized for recursion.
 */
template<>
class Matrix<0, 0, 0, 0>
{
  Matrix();
  Matrix(const Matrix&);
  Matrix& operator=(const Matrix&);

public:
  template<class E1, class E2, class Fcnl>
  static inline void assign2(E1&, const E2&, const Fcnl&) { }

  template<class E1, class E2, class Fcnl>
  static inline void assign(E1&, const E2&, const Fcnl&) { }

  template<class E>
  static inline bool all_elements2(const E&) { return true; }

  template<class E>
  static inline bool all_elements(const E&) { return true; }

  template<class E>
  static inline bool any_elements2(const E&) { return false; }

  template<class E>
  static inline bool any_elements(const E&) { return false; }

  template<class E>
  static inline XprNull trace(const E&) { return XprNull(); }
};


} // namespace meta

} // namespace tvmet

#endif /* TVMET_META_MATRIX_H */

// Local Variables:
// mode:C++
// End:
