/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Vector.h,v 1.16 2004/04/10 06:05:38 opetzold Exp $
 */

#ifndef TVMET_META_VECTOR_H
#define TVMET_META_VECTOR_H

#include <tvmet/xpr/Null.h>

namespace tvmet {

/* forwards */
template<class T, std::size_t Sz> class Vector;


namespace meta {


/**
 * \class Vector Vector.h "tvmet/meta/Vector.h"
 * \brief Meta %Vector class using expression templates
 */
template<std::size_t Sz, std::size_t Stride=0>
class Vector
{
  Vector();
  Vector(const Vector&);
  Vector& operator=(const Vector&);

private:
  enum {
    doIt = (Stride < (Sz-1)) ? 1 : 0		/**< recursive counter */
  };

public:
  /** Assign an expression expr using the functional fn. */
  template <class E1, class E2, class Fcnl>
  static inline
  void assign(E1& lhs, const E2& rhs, const Fcnl& fn) {
    fn.apply_on(lhs(Stride), rhs(Stride));
    meta::Vector<Sz * doIt, (Stride+1) * doIt>::assign(lhs, rhs, fn);
  }

  /** build the sum of the vector. */
  template<class E>
  static inline
  typename E::value_type
  sum(const E& e) {
    return e(Stride) + meta::Vector<Sz * doIt, (Stride+1) * doIt>::sum(e);
  }

  /** build the product of the vector. */
  template<class E>
  static inline
  typename NumericTraits<
    typename E::value_type
  >::sum_type
  product(const E& e) {
    return e(Stride) * meta::Vector<Sz * doIt, (Stride+1) * doIt>::product(e);
  }

  /** build the dot product of the vector. */
  template<class E1, class E2>
  static inline
  typename PromoteTraits<
    typename E1::value_type,
    typename E2::value_type
  >::value_type
  dot(const E1& lhs, const E2& rhs) {
    return lhs(Stride) * rhs(Stride)
      + meta::Vector<Sz * doIt, (Stride+1) * doIt>::dot(lhs, rhs);
  }

  /** check for all elements */
  template<class E>
  static inline
  bool
  all_elements(const E& e) {
    if(!e(Stride)) return false;
    return meta::Vector<Sz * doIt, (Stride+1) * doIt>::all_elements(e);
  }

  /** check for any elements */
  template<class E>
  static inline
  bool
  any_elements(const E& e) {
    if(e(Stride)) return true;
    return meta::Vector<Sz * doIt, (Stride+1) * doIt>::any_elements(e);
  }
};


/**
 * \class Vector<0,0> Vector.h "tvmet/meta/Vector.h"
 * \brief Meta %Vector Specialized for recursion
 */
template<>
class Vector<0,0>
{
  Vector();
  Vector(const Vector&);
  Vector& operator=(const Vector&);

public:
  template <class E1, class E2, class Fcnl>
  static inline void assign(E1&, const E2&, const Fcnl&) { }

  template<class E>
  static inline XprNull sum(const E&) { return XprNull(); }

  template<class E>
  static inline XprNull product(const E&) { return XprNull(); }

  template<class E1, class E2>
  static inline XprNull dot(const E1&, const E2&) { return XprNull(); }

  template<class E>
  static inline bool all_elements(const E&) { return true; }

  template<class E>
  static inline bool any_elements(const E&) { return false; }
};


} // namespace meta

} // namespace tvmet

#endif /* TVMET_META_VECTOR_H */

// Local Variables:
// mode:C++
// End:
